/*
  Unbound resource file for windows.  For use with windres
*/
#include "config.h"
#include "winver.h"

1 ICON "winrc/combined.ico"

1 VERSIONINFO
FILEVERSION	RSRC_PACKAGE_VERSION
PRODUCTVERSION  RSRC_PACKAGE_VERSION
FILEFLAGSMASK	0
FILEFLAGS	0
FILEOS		VOS__WINDOWS32
FILETYPE	VFT_APP
FILESUBTYPE	0
BEGIN
  BLOCK "StringFileInfo"
  BEGIN
    BLOCK "040904E4"
    BEGIN
      VALUE "CompanyName",	"NLnet Labs"
      VALUE "FileDescription",	"Unbound Service Install Util"
      VALUE "FileVersion",	PACKAGE_VERSION
      VALUE "InternalName",	"unbound-service-install"
      VALUE "OriginalFilename",	"unbound-service-install.exe"
      VALUE "ProductName",	"unbound"
      VALUE "ProductVersion",	PACKAGE_VERSION
      VALUE "LegalCopyright",	"(C) 2009 NLnet Labs. Source is BSD licensed."
    END
  END
  BLOCK "VarFileInfo"
  BEGIN
    /* English(409), windows ANSI codepage (1252) */
    VALUE "Translation", 0x409, 0x1252
  END
END

/* error message formats */
LANGUAGE 0x9,0x1
1 11 "winrc/gen_msg.bin"

/* vista administrator access, show UAC prompt */
1 RT_MANIFEST "winrc/vista_admin.manifest"

